/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#if !defined(_CRITSECT_H__INCLUDED_)
#define _CRITSECT_H__INCLUDED_

#ifdef _WIN32

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class CCritSect
{
public:
	CCritSect()
	{
		InitializeCriticalSection(&cs);
	}
	~CCritSect()
	{
		DeleteCriticalSection(&cs);
	}

	void Lock()
	{
		EnterCriticalSection(&cs);
	}
	void Unlock()
	{
		LeaveCriticalSection(&cs);
	}

	bool IsLocked()
	{
		return cs.LockCount > 0;
	}

protected:
	CRITICAL_SECTION cs;
};

#else // _WIN32

#include "SDL.h"
#include "SDL_thread.h"

class CCritSect
{
public:
	CCritSect()
	{
		cs = SDL_CreateMutex();
		locked = 0;
	}
	~CCritSect()
	{
		SDL_DestroyMutex( cs );
	}

	void Lock()
	{
		SDL_LockMutex( cs );
		locked = 1;
	}
	void Unlock()
	{
		locked = 0;
		SDL_UnlockMutex( cs );
	}

	bool IsLocked()
	{
		return (locked != 0);
	}

protected:
	SDL_mutex *cs;
	int locked;
};

#endif // _WIN32

#endif // !defined(_CRITSECT_H__INCLUDED_)
