#include <string.h>

#include "winlnxdefs.h"
#include "daedalus_math.h"

//---------- D3DXMATRIX

D3DXMATRIX::D3DXMATRIX()
{
}

D3DXMATRIX::D3DXMATRIX( const float *pIn )
{
   memcpy(m, pIn, 16*4);
}

D3DXMATRIX::D3DXMATRIX( const D3DMATRIX &pIn )
{
   memcpy(m, pIn.m, 16*4);
}

D3DXMATRIX::D3DXMATRIX( float _11, float _12, float _13, float _14,
		      float _21, float _22, float _23, float _24,
		      float _31, float _32, float _33, float _34,
		      float _41, float _42, float _43, float _44 )
{
   this->_11 = _11;
   this->_12 = _12;
   this->_13 = _13;
   this->_14 = _14;
   this->_21 = _21;
   this->_22 = _22;
   this->_23 = _23;
   this->_24 = _24;
   this->_31 = _31;
   this->_32 = _32;
   this->_33 = _33;
   this->_34 = _34;
   this->_41 = _41;
   this->_42 = _42;
   this->_43 = _43;
   this->_44 = _44;
}

float& D3DXMATRIX::operator () ( uint Row, uint Col )
{
   return m[Row][Col];
}

float  D3DXMATRIX::operator () ( uint Row, uint Col ) const
{
   return m[Row][Col];
}

D3DXMATRIX::operator float* ()
{
   return (float*)m;
}

D3DXMATRIX::operator const float* () const
{
   return (float*)m;
}

D3DXMATRIX& D3DXMATRIX::operator *= ( const D3DXMATRIX &pIn )
{
   D3DXMATRIX mTemp(*this);
   *this = mTemp*pIn;
}

D3DXMATRIX& D3DXMATRIX::operator += ( const D3DXMATRIX &pIn )
{
   D3DXMATRIX mTemp(*this);
   *this = mTemp+pIn;
}

D3DXMATRIX& D3DXMATRIX::operator -= ( const D3DXMATRIX &pIn )
{
   D3DXMATRIX mTemp(*this);
   *this = mTemp-pIn;
}

D3DXMATRIX& D3DXMATRIX::operator *= ( float f)
{
   for (int i=0; i<16; i++) ((float*)m)[i] *= f;
}

D3DXMATRIX& D3DXMATRIX::operator /= ( float f)
{
   for (int i=0; i<16; i++) ((float*)m)[i] /= f;
}

D3DXMATRIX D3DXMATRIX::operator + () const
{
   return *this;
}

D3DXMATRIX D3DXMATRIX::operator - () const
{
   D3DXMATRIX mTemp;
   for (int i=0; i<16; i++) ((float*)mTemp.m)[i] = -((float*)m)[i];
   return mTemp;
}

D3DXMATRIX D3DXMATRIX::operator * ( const D3DXMATRIX &pIn ) const
{
   D3DXMATRIX mTemp;
   for (int i=0; i<4; i++)
     for (int j=0; j<4; j++)
       mTemp.m[i][j] = m[i][0]*pIn.m[0][j] +
                       m[i][1]*pIn.m[1][j] +
                       m[i][2]*pIn.m[2][j] +
                       m[i][3]*pIn.m[3][j];
   return mTemp;
}

D3DXMATRIX D3DXMATRIX::operator + ( const D3DXMATRIX &pIn ) const
{
   D3DXMATRIX mTemp;
   for (int i=0; i<16; i++)
     ((float*)mTemp.m)[i] = ((float*)m)[i] + ((float*)pIn.m)[i];
   return mTemp;
}

D3DXMATRIX D3DXMATRIX::operator - ( const D3DXMATRIX &pIn ) const
{
   D3DXMATRIX mTemp;
   for (int i=0; i<16; i++)
     ((float*)mTemp.m)[i] = ((float*)m)[i] - ((float*)pIn.m)[i];
   return mTemp;
}

/*
    D3DXMATRIX operator * ( float ) const;
    D3DXMATRIX operator / ( float ) const;
    friend D3DXMATRIX operator * ( float, const D3DXMATRIX & );
    bool operator == ( const D3DXMATRIX & ) const;
    bool operator != ( const D3DXMATRIX & ) const;
*/

//---------- VECTOR3

D3DXVECTOR3::D3DXVECTOR3()
{
}

/*
	D3DXVECTOR3( const float *f );
	D3DXVECTOR3( const D3DVECTOR3 &v );
	D3DXVECTOR3( float _x, float _y, float _z );

	// casting
	inline operator float* ();
	inline operator const float* () const;

	// assignment operators
	inline D3DXVECTOR3& operator += ( const D3DXVECTOR3 &op );
	inline D3DXVECTOR3& operator -= ( const D3DXVECTOR3 &op );
	inline D3DXVECTOR3& operator *= ( float op );
        inline D3DXVECTOR3& operator /= ( float op );

	// unary operators
	inline D3DXVECTOR3 operator + () const;
	inline D3DXVECTOR3 operator - () const;

	// binary operators
        inline D3DXVECTOR3 operator + ( const D3DXVECTOR3 &op ) const;
	inline D3DXVECTOR3 operator - ( const D3DXVECTOR3 &op ) const;
	inline D3DXVECTOR3 operator * ( float op ) const;
	inline D3DXVECTOR3 operator / ( float op ) const;


	friend D3DXVECTOR3 operator * ( float, const D3DXVECTOR3& );

	inline bool operator == ( const D3DXVECTOR3 &op ) const;
	inline bool operator != ( const D3DXVECTOR3 &op ) const;
*/

//---------- VECTOR4

D3DXVECTOR4::D3DXVECTOR4()
{
}

/*
	D3DXVECTOR4( const float *f );
	D3DXVECTOR4( const D3DVECTOR4 &v );
	D3DXVECTOR4( float _x, float _y, float _z, float _w );

	// casting
        inline operator float* ();
	inline operator const float* () const;

	// assignment operators
	inline D3DXVECTOR4& operator += ( const D3DXVECTOR4 &op );
	inline D3DXVECTOR4& operator -= ( const D3DXVECTOR4 &op );
	inline D3DXVECTOR4& operator *= ( float op );
	inline D3DXVECTOR4& operator /= ( float op );

	// unary operators
	inline D3DXVECTOR4 operator + () const;
	inline D3DXVECTOR4 operator - () const;

	// binary operators
	inline D3DXVECTOR4 operator + ( const D3DXVECTOR4 &op ) const;
	inline D3DXVECTOR4 operator - ( const D3DXVECTOR4 &op ) const;
	inline D3DXVECTOR4 operator * ( float op ) const;
	inline D3DXVECTOR4 operator / ( float op ) const;

	friend D3DXVECTOR4 operator * ( float, const D3DXVECTOR4& );

	inline bool operator == ( const D3DXVECTOR4 &op ) const;
	inline bool operator != ( const D3DXVECTOR4 &op ) const;
*/

//---------- OTHER

D3DXMATRIX* D3DXMatrixTranspose(
				D3DXMATRIX* pOut,
				CONST D3DXMATRIX* pM
				)
{
   pOut->_11 = pM->_11;
   pOut->_12 = pM->_21;
   pOut->_13 = pM->_31;
   pOut->_14 = pM->_41;
   pOut->_21 = pM->_12;
   pOut->_22 = pM->_22;
   pOut->_23 = pM->_32;
   pOut->_24 = pM->_42;
   pOut->_31 = pM->_13;
   pOut->_32 = pM->_23;
   pOut->_33 = pM->_33;
   pOut->_34 = pM->_43;
   pOut->_41 = pM->_14;
   pOut->_42 = pM->_24;
   pOut->_43 = pM->_34;
   pOut->_44 = pM->_44;
   return pOut;
}

D3DXVECTOR4 WINAPI D3DXVec3Transform(
				     D3DXVECTOR4 *pOut,
				     CONST D3DXVECTOR3 *pV,
				     CONST D3DXMATRIX *pM
				     )
{
   pOut->x = pV->x*pM->_11 + pV->y*pM->_21 + pV->z*pM->_31 + pM->_41;
   pOut->y = pV->x*pM->_12 + pV->y*pM->_22 + pV->z*pM->_32 + pM->_42;
   pOut->z = pV->x*pM->_13 + pV->y*pM->_23 + pV->z*pM->_33 + pM->_43;
   pOut->w = pV->x*pM->_14 + pV->y*pM->_24 + pV->z*pM->_34 + pM->_44;
   return *pOut;
}
